/*
 * xhtmlundo.c
 * xhtmlite
 *
 * by Doug Wolfe
 * doug@wolfescience.com 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc.,  59 Temple Place - Suite 330, Cambridge, MA 02139, USA.
 *
 */

#include <gtk/gtk.h>
#include "xhtmlundo.h"

/*TODO: Fix Bug: Spaces are not properly undoed and redoed*/
static void
xhtmlite_on_insert_text (GtkTextBuffer *textbuffer, GtkTextIter *iter,
                                            gchar *text, gint position, gpointer user_data) 
{
	int undo = GPOINTER_TO_INT(g_object_get_data (G_OBJECT(textbuffer), "process-undo"));
	if(undo)
	{
		xhtmlite_undo_add_redo(textbuffer, iter, NULL, text, NULL, XHTMLITE_UNDO_INSERT);
	}
	else
	{
		xhtmlite_undo_add_undo(textbuffer, iter, NULL, text, NULL, XHTMLITE_UNDO_INSERT);
	}
} 

static void
xhtmlite_on_insert_pixbuf (GtkTextBuffer *textbuffer, GtkTextIter *start, GdkPixbuf *pixbuf, gpointer user_data)
{
	int undo = GPOINTER_TO_INT(g_object_get_data (G_OBJECT(textbuffer), "process-undo"));
	if(undo)
	{
		xhtmlite_undo_add_redo(textbuffer, start, NULL, NULL, NULL, XHTMLITE_UNDO_INSERT_PIXBUF);
	}
	else
	{
		xhtmlite_undo_add_undo(textbuffer, start, NULL, NULL, NULL, XHTMLITE_UNDO_INSERT_PIXBUF);
	}	
}

static void
xhtmlite_on_delete_range (GtkTextBuffer *textbuffer, GtkTextIter *start, GtkTextIter *end, gpointer user_data)
{
	int undo = GPOINTER_TO_INT(g_object_get_data (G_OBJECT(textbuffer), "process-undo"));
	if(undo)
	{
		xhtmlite_undo_add_redo(textbuffer, start, end, NULL, NULL, XHTMLITE_UNDO_DELETE);	
	}
	else
	{
		xhtmlite_undo_add_undo(textbuffer, start, end, NULL, NULL, XHTMLITE_UNDO_DELETE);	
	}
}

static void
xhtmlite_undo_set_buttons(GtkTextBuffer *textbuffer)
{
	GtkWidget *button = g_object_get_data (G_OBJECT(textbuffer), "undo_button");
	if(button != NULL)
	{
		gtk_widget_set_sensitive (button, xhtmlite_undo_has_undo(textbuffer));
	}

	button = g_object_get_data (G_OBJECT(textbuffer), "redo_button");
	if(button != NULL)
	{
		gtk_widget_set_sensitive (button, xhtmlite_undo_has_redo(textbuffer));
	}
}

void xhtmlite_undo_assign_buttons(GtkTextBuffer *textbuffer, GtkWidget *undo_button, GtkWidget *redo_button)
{
	g_object_set_data (G_OBJECT(textbuffer), "undo_button", undo_button);
	g_object_set_data (G_OBJECT(textbuffer), "redo_button", redo_button);
	xhtmlite_undo_set_buttons(textbuffer);	
}

void xhtmlite_undo_initialize_undos(GtkTextBuffer *textbuffer)
{
	GList *undo_list = NULL;
	g_object_set_data (G_OBJECT(textbuffer), "undo", undo_list);
	GList *redo_list = NULL;	
	g_object_set_data (G_OBJECT(textbuffer), "redo", redo_list);
	g_object_set_data (G_OBJECT(textbuffer), "process-undo", GINT_TO_POINTER(0));
	g_object_set_data (G_OBJECT(textbuffer), "process-redo", GINT_TO_POINTER(0));
	
		g_signal_connect ((gpointer) textbuffer, "insert_text",
				G_CALLBACK (xhtmlite_on_insert_text),
				NULL);
		g_signal_connect ((gpointer) textbuffer, "insert_pixbuf",
				G_CALLBACK (xhtmlite_on_insert_pixbuf),
				NULL);
		g_signal_connect ((gpointer) textbuffer, "delete_range",
				G_CALLBACK (xhtmlite_on_delete_range),
				NULL);
}

void xhtmlite_undo_clear(GtkTextBuffer *textbuffer)
{
	xhtmlite_undo_clear_undos(textbuffer);	
	xhtmlite_undo_clear_redos(textbuffer);
	xhtmlite_undo_set_buttons(textbuffer);
}

void xhtmlite_undo_clear_undos(GtkTextBuffer *textbuffer)
{
	//for new files or opened files
	GList *list = g_object_get_data (G_OBJECT(textbuffer), "undo");
	if(list != NULL)
	{
		while(g_list_length(list)>0)
		{
			xhtmlite_undo_store *undo_data = g_list_nth_data (list, 0);
			list = g_list_remove(list, undo_data);
		}
	}
}

void xhtmlite_undo_clear_redos(GtkTextBuffer *textbuffer)
{
	//for new files or opened files
	GList *list = g_object_get_data (G_OBJECT(textbuffer), "redo");
	if(list != NULL)
	{
		while(g_list_length(list)>0)
		{
			xhtmlite_undo_store *redo_data = g_list_nth_data (list, 0);
			list = g_list_remove(list, redo_data);
		}
	}
}

static xhtmlite_undo_store*
xhtmlite_undo_add_do (GtkTextBuffer *textbuffer, GtkTextIter *start, GtkTextIter *end, gchar *text, GtkTextTag *tag, gint undo_type)
{
	xhtmlite_undo_store *data = malloc(sizeof(xhtmlite_undo_store));
	data->undo_type = undo_type;
	switch (undo_type)
	{
		case(XHTMLITE_UNDO_INSERT):
			data->start = gtk_text_iter_get_offset (start);
			printf("data->start:%d\n", data->start);
			data->end = gtk_text_iter_get_offset (start) + g_str_get_length(text,-1);
			printf("data->end:%d\n", data->end);
			data->text = g_strdup(text);
			printf("data->text:%s\n", data->text);
			break;
		case(XHTMLITE_UNDO_INSERT_PIXBUF):	
			data->start = gtk_text_iter_get_offset (start);
			printf("data->start:%d\n", data->start);
			data->end = gtk_text_iter_get_offset (start)+1;
			printf("data->end:%d\n", data->end);
			GtkTextIter end_iter;
			gtk_text_buffer_get_iter_at_offset (textbuffer, &end_iter, data->end);
			data->text = xhtmlite_write_range(textbuffer, start, &end_iter);
			printf("data->text:%s\n", data->text);
			break;
		case(XHTMLITE_UNDO_DELETE):
			data->start = gtk_text_iter_get_offset (start);
			printf("data->start:%d\n", data->start);
			data->end = gtk_text_iter_get_offset (end);
			printf("data->end:%d\n", data->end);
			data->text = xhtmlite_write_range(textbuffer, start, end);
			printf("data->text:%s\n", data->text);
			break;
		case(XHTMLITE_UNDO_ADD_TAG):
			data->start = gtk_text_iter_get_offset (start);
			printf("data->start:%d\n", data->start);
			data->end = gtk_text_iter_get_offset (end);
			printf("data->end:%d\n", data->end);
			data->tag = tag;
			printf("data->tag:%s\n", tag->name);
			break;			
		case(XHTMLITE_UNDO_REMOVE_TAG):	
			data->start = gtk_text_iter_get_offset (start);
			printf("data->start:%d\n", data->start);
			data->end = gtk_text_iter_get_offset (end);
			printf("data->end:%d\n", data->end);
			data->tag = tag;
			printf("data->tag:%s\n", tag->name);
			break;	
		default:
			printf("xhtmlite_undo_add_undo: other type\n");
	}
	return data;
}
void xhtmlite_undo_add_undo (GtkTextBuffer *textbuffer, GtkTextIter *start, GtkTextIter *end, gchar *text, GtkTextTag *tag, gint undo_type)
{
	printf("xhtmlite_undo_add_undo\n");
	GList *list = g_object_get_data (G_OBJECT(textbuffer), "undo");
	
	xhtmlite_undo_store *data = xhtmlite_undo_add_do(textbuffer, start, end, text, tag, undo_type);
	
	list = g_list_prepend (list, data);
	g_object_set_data (G_OBJECT(textbuffer), "undo", list);
	//xhtmlite_undo_clear_redos(textbuffer);
	xhtmlite_undo_set_buttons(textbuffer);
}

void xhtmlite_undo_add_redo (GtkTextBuffer *textbuffer, GtkTextIter *start, GtkTextIter *end, gchar *text, GtkTextTag *tag, gint undo_type)
{
	printf("xhtmlite_undo_add_redo\n");
	GList *list = g_object_get_data (G_OBJECT(textbuffer), "redo");
	
	xhtmlite_undo_store *data = xhtmlite_undo_add_do(textbuffer, start, end, text, tag, undo_type);
	
	list = g_list_prepend (list, data);
	g_object_set_data (G_OBJECT(textbuffer), "redo", list);
	xhtmlite_undo_set_buttons(textbuffer);
}

int xhtmlite_undo_has_undo (GtkTextBuffer *textbuffer)
{
	printf("xhtmlite has undo:");
	GList *list = g_object_get_data (G_OBJECT(textbuffer), "undo");
	if(g_list_length(list)>0)
	{
		printf("TRUE\n");
		return 1;
	}
	printf("FALSE\n");
	return 0;   
}

int xhtmlite_undo_has_redo (GtkTextBuffer *textbuffer)
{
	printf("xhtmlite has redo:");
	GList *list = g_object_get_data (G_OBJECT(textbuffer), "redo");
	if(g_list_length(list)>0)
	{
		printf("TRUE\n");
		return 1;
	}
	printf("FALSE\n");
	return 0;   
}

void xhtmlite_undo_undo (GtkTextBuffer *textbuffer)
{
	printf("Start UNDO\n");
	
	//		g_signal_handlers_block_matched (G_OBJECT(textbuffer),
	//			     G_SIGNAL_MATCH_FUNC, 0, 0,
	//			     NULL, G_CALLBACK(xhtmlite_on_buffer_changed), NULL);
	g_object_set_data (G_OBJECT(textbuffer), "process-undo", GINT_TO_POINTER(1));
	GList *list = g_object_get_data (G_OBJECT(textbuffer), "undo");
	xhtmlite_undo_store * data = g_list_nth_data (list, 0);
	GtkTextIter start, end;
		
	switch (data->undo_type)
	{
		case(XHTMLITE_UNDO_INSERT):
			gtk_text_buffer_get_iter_at_offset (textbuffer, &start, data->start);
			gtk_text_buffer_get_iter_at_offset (textbuffer, &end, data->end);
			gtk_text_buffer_delete(textbuffer, &start, &end);
			gtk_text_buffer_get_iter_at_offset (textbuffer, &start, data->start);
			gtk_text_buffer_select_range (textbuffer,&start,&start);
			break;
		case(XHTMLITE_UNDO_INSERT_PIXBUF):	
			gtk_text_buffer_get_iter_at_offset (textbuffer, &start, data->start);
			gtk_text_buffer_get_iter_at_offset (textbuffer, &end, data->end);
			gtk_text_buffer_delete(textbuffer, &start, &end);
			break;
		case(XHTMLITE_UNDO_DELETE):
			gtk_text_buffer_get_iter_at_offset (textbuffer, &start, data->start);
			int end_offset = xhtmlite_insert (textbuffer, data->text, &start, FALSE);
			gtk_text_buffer_get_iter_at_offset (textbuffer, &end, end_offset);
			gtk_text_buffer_select_range (textbuffer,&start,&start);
			break;
		case(XHTMLITE_UNDO_ADD_TAG):
			gtk_text_buffer_get_iter_at_offset (textbuffer, &start, data->start);
			gtk_text_buffer_get_iter_at_offset (textbuffer, &end, data->end);
			gtk_text_buffer_remove_tag (textbuffer, data->tag, &start, &end);
			xhtmlite_undo_add_redo(textbuffer, &start, &end, NULL, data->tag, XHTMLITE_UNDO_REMOVE_TAG);
			gtk_text_buffer_select_range (textbuffer,&start,&end);			
			break;
		case(XHTMLITE_UNDO_REMOVE_TAG):	
			gtk_text_buffer_get_iter_at_offset (textbuffer, &start, data->start);
			gtk_text_buffer_get_iter_at_offset (textbuffer, &end, data->end);
			gtk_text_buffer_apply_tag (textbuffer, data->tag, &start, &end);
			xhtmlite_undo_add_redo(textbuffer, &start, &end, NULL, data->tag, XHTMLITE_UNDO_ADD_TAG);
			gtk_text_buffer_select_range (textbuffer,&start,&end);			
			break;
		default:
			printf("xhtmlite_undo_undo: other type\n");
	}
	list = g_list_remove(list, data);
	g_object_set_data (G_OBJECT(textbuffer), "undo", list);
	xhtmlite_undo_set_buttons(textbuffer);
	g_object_set_data (G_OBJECT(textbuffer), "process-undo", GINT_TO_POINTER(0));
//		g_signal_handlers_unblock_matched (G_OBJECT(textbuffer),
	//			       G_SIGNAL_MATCH_FUNC, 0, 0, NULL,
	//			       G_CALLBACK(xhtmlite_on_buffer_changed), NULL);	
}

void xhtmlite_undo_redo (GtkTextBuffer *textbuffer)
{
	printf("Start REDO\n");
	g_object_set_data (G_OBJECT(textbuffer), "process-redo", GINT_TO_POINTER(1));
	GList *list = g_object_get_data (G_OBJECT(textbuffer), "redo");
	xhtmlite_undo_store * data = g_list_nth_data (list, 0);
	GtkTextIter start, end;
	
	switch (data->undo_type)
	{
		case(XHTMLITE_UNDO_INSERT):
			gtk_text_buffer_get_iter_at_offset (textbuffer, &start, data->start);
			gtk_text_buffer_get_iter_at_offset (textbuffer, &end, data->end);
			gtk_text_buffer_delete(textbuffer, &start, &end);
			gtk_text_buffer_get_iter_at_offset (textbuffer, &start, data->start);
			gtk_text_buffer_select_range (textbuffer,&start,&start);
			break;
		case(XHTMLITE_UNDO_INSERT_PIXBUF):	
			gtk_text_buffer_get_iter_at_offset (textbuffer, &start, data->start);
			gtk_text_buffer_get_iter_at_offset (textbuffer, &end, data->start + 1);
			gtk_text_buffer_delete(textbuffer, &start, &end);
			break;
		case(XHTMLITE_UNDO_DELETE):
			gtk_text_buffer_get_iter_at_offset (textbuffer, &start, data->start);
			gtk_text_buffer_select_range (textbuffer,&start,&start);
			xhtmlite_insert (textbuffer, data->text, &start, FALSE);
			break;
		case(XHTMLITE_UNDO_ADD_TAG):
			gtk_text_buffer_get_iter_at_offset (textbuffer, &start, data->start);
			gtk_text_buffer_get_iter_at_offset (textbuffer, &end, data->end);
			gtk_text_buffer_remove_tag (textbuffer, data->tag, &start, &end);
			xhtmlite_undo_add_undo(textbuffer, &start, &end, NULL, data->tag, XHTMLITE_UNDO_REMOVE_TAG);
			gtk_text_buffer_select_range (textbuffer,&start,&end);			
			break;
		case(XHTMLITE_UNDO_REMOVE_TAG):	
			gtk_text_buffer_get_iter_at_offset (textbuffer, &start, data->start);
			gtk_text_buffer_get_iter_at_offset (textbuffer, &end, data->end);
			gtk_text_buffer_apply_tag (textbuffer, data->tag, &start, &end);
			xhtmlite_undo_add_undo(textbuffer, &start, &end, NULL, data->tag, XHTMLITE_UNDO_ADD_TAG);
			gtk_text_buffer_select_range (textbuffer,&start,&end);			
			break;
		default:
			printf("xhtmlite_undo_redo: other type\n");
	}
	list = g_list_remove(list, data);
	g_object_set_data (G_OBJECT(textbuffer), "redo", list);
	xhtmlite_undo_set_buttons(textbuffer);
	g_object_set_data (G_OBJECT(textbuffer), "process-redo", GINT_TO_POINTER(0));
}
